
// http://codesynthesis.com/~boris/blog/2010/05/17/parsing-cxx-with-gcc-plugin-part-3/
// http://www.codesynthesis.com/~boris/data/gcc-plugin/plugin-3.cxx

// see <gcc-sources>/gcc/testsuite/g++.dg/plugin

// GCC header includes to get the parse tree declarations. The order
// is important and doesn't follow any kind of logic.

#include <stdlib.h>
#include <gmp.h>

#include <cstdlib> // Include before GCC poison some declarations.

#include "gcc-plugin.h"

#include "config.h"
#include "system.h"
#include "coretypes.h"
#include "tree.h"
#include "intl.h"

#include "tm.h"

#include "cp/cp-tree.h"
#include "c-common.h"
#include "c-pragma.h"

// #include "diagnostic.h"

#include <set>
#include <string>
#include <iostream>

#include "support.h"
using namespace NAMESPACE;

using namespace std;

int plugin_is_GPL_compatible;

THtmlOutput io;

void print_decl (tree decl);

string decl_scope (tree decl)
{
  string s, tmp;

  for (tree scope (CP_DECL_CONTEXT (decl));
       scope != global_namespace;
       scope = CP_DECL_CONTEXT (scope))
  {
    if (TREE_CODE (scope) == RECORD_TYPE)
      scope = TYPE_NAME (scope);

    tree id (DECL_NAME (scope));

    tmp = "::";
    tmp += (id != 0 ? IDENTIFIER_POINTER (id) : "<unnamed>");
    tmp += s;
    s.swap (tmp);
  }

  return s;
}

enum access_spec
{
  public_, protected_, private_
};

const char* access_spec_str[] =
{
  "public", "protected", "private"
};

void print_class (tree type)
{
  type = TYPE_MAIN_VARIANT (type);

  tree decl = TYPE_NAME (type);
  tree id  = DECL_NAME (decl);
  const char * name = IDENTIFIER_POINTER (id);

  io.StyledText ("class " + decl_scope (decl) + name, class_style);
  if (!COMPLETE_TYPE_P (type)) io.Put (";");

  io.StyledText (string (" // at ") + DECL_SOURCE_FILE (decl) +  ":" + IntToStr (DECL_SOURCE_LINE (decl)) , comment_style);
  io.PutEol ();

  // We are done if this is an incomplete
  // class declaration.
  //
  if (COMPLETE_TYPE_P (type))
  {

     io.StyledText ("{", class_style);
     io.PutEol ();
     io.IncIndent ();

     // Traverse base information.
     //
     tree biv (TYPE_BINFO (type));
     size_t n (biv ? BINFO_N_BASE_BINFOS (biv) : 0);

     for (size_t i (0); i < n; i++)
     {
       tree bi (BINFO_BASE_BINFO (biv, i));

       // Get access specifier.
       //
       access_spec a (public_);

       if (BINFO_BASE_ACCESSES (biv))
       {
         tree ac (BINFO_BASE_ACCESS (biv, i));

         if (ac == 0 || ac == access_public_node)
           a = public_;
         else if (ac == access_protected_node)
           a = protected_;
         else
           a = private_;
       }

       bool virt (BINFO_VIRTUAL_P (bi));
       tree b_type (TYPE_MAIN_VARIANT (BINFO_TYPE (bi)));
       tree b_decl (TYPE_NAME (b_type));
       tree b_id (DECL_NAME (b_decl));
       const char* b_name (IDENTIFIER_POINTER (b_id));

       // cerr << "\t" << access_spec_str[a] << (virt ? " virtual" : "")
       //      << " base " << decl_scope (b_decl) << "::" << b_name << endl;
     }

     // Traverse members.
     //
     for (tree d (TYPE_FIELDS (type)); d != 0; d = TREE_CHAIN (d))
     {
       switch (TREE_CODE (d))
       {
       case TYPE_DECL:
         {
           if (!DECL_SELF_REFERENCE_P (d))
             print_decl (d);
           break;
         }
       case FIELD_DECL:
         {
           if (!DECL_ARTIFICIAL (d))
             print_decl (d);
           break;
         }
       default:
         {
           print_decl (d);
           break;
         }
       }
     }

     for (tree d (TYPE_METHODS (type)); d != 0; d = TREE_CHAIN (d))
     {
       if (!DECL_ARTIFICIAL (d))
         print_decl (d);
     }

     io.DecIndent ();
     io.StyledText ("};", class_style);
     io.PutEol ();
  }
}

void print_decl (tree decl)
{
  tree type = TREE_TYPE (decl);
  int dc = TREE_CODE (decl);
  int tc = 0;

  if (type)
  {
    tc = TREE_CODE (type);

    if (dc == TYPE_DECL && tc == RECORD_TYPE)
    {
      // If DECL_ARTIFICIAL is true this is a class
      // declaration. Otherwise this is a typedef.
      if (DECL_ARTIFICIAL (decl))
      {
        print_class (type);
        return;
      }
    }
  }

  tree id = DECL_NAME (decl);
  string name = id ? IDENTIFIER_POINTER (id) : "<unnamed>";

  io.StyledText (/* + decl_scope (decl) + "::" + */ name, variable_style);

  io.StyledText (string (" // ") + get_tree_code_name((tree_code)dc), property_style);

  if (type)
    io.StyledText (string (", type ") + get_tree_code_name((tree_code)tc), property_style);

  io.StyledText (string (" // at ") + DECL_SOURCE_FILE (decl) + ":" + IntToStr (DECL_SOURCE_LINE (decl)), comment_style);
  io.PutEol ();
}

void collect (tree ns)
{
  tree decl;
  cp_binding_level * level = NAMESPACE_LEVEL (ns);

  // Collect declarations.
  for (decl = level->names; decl != 0; decl = TREE_CHAIN (decl))
  {
    if (DECL_IS_BUILTIN (decl))
       continue;

    print_decl (decl);
  }

  // Traverse namespaces.
  for(decl = level->namespaces; decl != 0; decl = TREE_CHAIN (decl))
  {
    if (DECL_IS_BUILTIN (decl))
      continue;

    io.StyledText ("namespace " /*  + decl_scope (decl) + IDENTIFIER_POINTER (decl) */, namespace_style);
    io.PutEol ();
    io.StyledText ("{", namespace_style);
    io.PutEol ();
    io.IncIndent ();

    collect (decl);

    io.DecIndent ();
    io.StyledText ("}", namespace_style);
    io.PutEol ();
  }

  // print_decl (ns);
}

extern "C" void gate_callback (void*, void*)
{
  // If there were errors during compilation,
  // let GCC handle the exit.
  if (errorcount || sorrycount)
    return;

  int result = 0;

  // Process AST. Issue diagnostics and set result  to 1 in case of an error.
  cerr << "processing " << main_input_filename << endl;

  io.Open ("output/output.html");
  io.Head ("");
  io.PutLn(string ("<h1>") + "pseudo code from " + main_input_filename + "</h1>");
  io.PutLn("<pre>");

  // set_error_handlers (); // ?!
  collect (global_namespace);

  io.PutLn("</pre>");
  io.Tail ();
  io.WriteText ();

  exit (result);
}

extern "C" int plugin_init (plugin_name_args * info, plugin_gcc_version * version)
{
  int result = 0;

  cerr << "starting " << info->base_name << endl;

  // Parse options if any.

  // Disable assembly output.
  asm_file_name = HOST_BIT_BUCKET;

  // Register callbacks.
  register_callback (info->base_name,
                     PLUGIN_OVERRIDE_GATE,
                     &gate_callback,
                     0);

  return result;
}
